#include <Arduino.h>
#include <SoftwareSerial.h>
#include <TinyGPSPlus.h>
#include <Adafruit_LIS3MDL.h>
#include <RF24.h>
#include <printf.h>
#include <math.h>
#include <LiquidCrystal_I2C.h>
#include <Servo.h>

// front right
#define AHBridgeEnA 13
#define AHBridgeIn1 12
#define AHBridgeIn2 11

// back right
#define AHBridgeIn3 10
#define AHBridgeIn4 9
#define AHBridgeEnB 8

// front left
#define BHBridgeEnA 7
#define BHBridgeIn1 6
#define BHBridgeIn2 5

// back left
#define BHBridgeIn3 4
#define BHBridgeIn4 3
#define BHBridgeEnB 2

#define gpsPinRx 30
#define gpsPinTx 31

#define TURNING_TIME_MS 0
// #define DEBUG
// #define POLARITY_DEBUG
// in ft
#define ROBOT_LENGTH 1.5
#define ROBOT_WIDTH 1.0

// how much the servo will move in degrees when command comes from remote
#define SERVO_ADJUSTMENT_FACTOR 20

// SoftwareSerial ss(gpsPinRx, gpsPinTx);
TinyGPSPlus gps;
Adafruit_LIS3MDL compass;
RF24 radio(49, 53); // CE 49, CSN 53
LiquidCrystal_I2C lcd(0x27, 16, 4);

static const byte RADIO_ADDRESS[6] = "00001";
sensors_event_t event;

enum class RobotDirection {
    LEFT,
    RIGHT,
    FORWARD,
    BACKWARD
};

enum class CardinalDirection {
  NORTH = 0,
  SOUTH = 180,
  EAST = 90,
  WEST = 270
};

Servo serv;

static const int HIGH_POL = 1;
static const int LOW_POL = 0;

// volleyball court coordinates
// bottom left corner
static const double SOUTHWEST_LAT = 40.00188;
static const double SOUTHWEST_LONG = 105.26248;
// top right corner
static const double NORTHEAST_LAT = 40.00200;
static const double NORTHEAST_LONG = 105.26240;

struct DirectionRelation
{
  int Aport;
  int Bport;
  DirectionRelation(const int InAPort = LOW_POL, const int InBPort = HIGH_POL)
  {
    Aport = InAPort;
    Bport = InBPort;
  }
};

DirectionRelation InverseRelation(HIGH_POL, LOW_POL);

class Wheel
{
private:
  RobotDirection current_direction = RobotDirection::FORWARD;
  DirectionRelation current_polarities;
  const DirectionRelation forward_polarities;
  int pin1;
  int pin2;
  int analogPin;

  void sendPolarities()
  {
    digitalWrite(pin1, current_polarities.Aport);
    digitalWrite(pin2, current_polarities.Bport);
#ifdef POLARITY_DEBUG
    Serial.print("Sent polarities to: ");
    Serial.print(pin1);
    Serial.print(", ");
    Serial.println(pin2);
    Serial.print("values: ");
    Serial.print(current_polarities.Aport);
    Serial.print(", ");
    Serial.println(current_polarities.Bport);
#endif
  }

public:
  Wheel(int analogPin, int pin1, int pin2, const DirectionRelation relativePolarities = {})
  {
    forward_polarities = relativePolarities;
    current_polarities = relativePolarities;
    this->pin1 = pin1;
    this->pin2 = pin2;
    this->analogPin = analogPin;

    sendPolarities();
  }

  void startMotor()
  {
    analogWrite(analogPin, 255);
  }

  void stopMotor()
  {
    analogWrite(analogPin, 0);
  }

  void reverseDirection()
  {
    current_polarities.Aport = !forward_polarities.Aport;
    current_polarities.Bport = !forward_polarities.Bport;
    sendPolarities();
  }

  void forwardDirection()
  {
    current_polarities.Aport = forward_polarities.Aport;
    current_polarities.Bport = forward_polarities.Bport;
    sendPolarities();
  }
};

Wheel FrontRightWheel(AHBridgeEnA, AHBridgeIn1, AHBridgeIn2, InverseRelation);
Wheel BackRightWheel(AHBridgeEnB, AHBridgeIn3, AHBridgeIn4, InverseRelation); // wrong ways

Wheel FrontLeftWheel(BHBridgeEnA, BHBridgeIn3, BHBridgeIn4, InverseRelation);
Wheel BackLeftWheel(BHBridgeEnB, BHBridgeIn1, BHBridgeIn2, InverseRelation); // wrong ways

void fullSteamAhead();
void rightTurn();
void leftTurn();

static const uint32_t GPSBaud = 9600;

#define RADIUS_OF_EARTH 6371000 // Radius of the Earth in meters
#define METERS_TO_FEET 3.28084  // Conversion factor for meters to feet

// Function to calculate distance between two points using Haversine formula
float haversine(float lat1, float lon1, float lat2, float lon2)
{
  float phi1 = radians(lat1);
  float phi2 = radians(lat2);
  float delta_phi = radians(lat2 - lat1);
  float delta_lambda = radians(lon2 - lon1);

  float a = sin(delta_phi / 2) * sin(delta_phi / 2) +
            cos(phi1) * cos(phi2) *
                sin(delta_lambda / 2) * sin(delta_lambda / 2);
  float c = 2 * atan2(sqrt(a), sqrt(1 - a));

  float distance = RADIUS_OF_EARTH * c;
  return distance;
}

void setup()
{
  // setup radio
  printf_begin();
  Serial.begin(9600);
  lcd.init();
  lcd.backlight();
  lcd.setCursor(0, 0);
  lcd.print("Initializing");
  lcd.setCursor(0, 1);
  lcd.print("Radio...");

  if (!radio.begin())
  {
    radio.printDetails();
    Serial.println("Radio setup failed...");
    if (!radio.isChipConnected())
    {
      Serial.println("Check the radio wiring, I don't think it's connected!");
    }
    while (1)
      ;
  }

  lcd.clear();
  lcd.setCursor(0,0);
  lcd.print("Radio ready!");

  delay(200); // let radio take settings

  radio.openReadingPipe(1, RADIO_ADDRESS);
  radio.startListening();
  // -------
  #ifdef DEBUG
    radio.printPrettyDetails();
  #endif

  // setup compass
  if (!compass.begin_I2C())
  {
    Serial.println("Failed to find LIS3MDL, did you wire to SDA and SCL correctly?");
      lcd.setCursor(0,0);
      lcd.print("Compass failure");
      lcd.setCursor(0,1);
      lcd.print("Check step 10");
    while (1)
      ;
  }

  compass.setOperationMode(LIS3MDL_CONTINUOUSMODE);
  compass.setDataRate(LIS3MDL_DATARATE_155_HZ);
  compass.setRange(LIS3MDL_RANGE_8_GAUSS);
  // -------

  // setup pins
  pinMode(AHBridgeEnA, OUTPUT);
  pinMode(AHBridgeIn1, OUTPUT);
  pinMode(AHBridgeIn2, OUTPUT);
  pinMode(AHBridgeIn3, OUTPUT);
  pinMode(AHBridgeIn4, OUTPUT);
  pinMode(AHBridgeEnB, OUTPUT);

  pinMode(BHBridgeEnA, OUTPUT);
  pinMode(BHBridgeIn1, OUTPUT);
  pinMode(BHBridgeIn2, OUTPUT);
  pinMode(BHBridgeIn3, OUTPUT);
  pinMode(BHBridgeIn4, OUTPUT);
  pinMode(BHBridgeEnB, OUTPUT);

  serv.attach(A0, 500, 2500);
  serv.write(40);
}

void fullSteamAhead()
{
  FrontLeftWheel.forwardDirection();
  FrontLeftWheel.startMotor();

  FrontRightWheel.forwardDirection();
  FrontRightWheel.startMotor();

  BackLeftWheel.forwardDirection();
  BackLeftWheel.startMotor();

  BackRightWheel.forwardDirection();
  BackRightWheel.startMotor();
}

void stopAllMotors()
{
  FrontLeftWheel.stopMotor();
  FrontRightWheel.stopMotor();
  BackLeftWheel.stopMotor();
  BackRightWheel.stopMotor();
}

void rightTurn()
{
  stopAllMotors();
  BackRightWheel.reverseDirection();
  BackRightWheel.startMotor();

  FrontRightWheel.reverseDirection();
  FrontRightWheel.startMotor();

  FrontLeftWheel.forwardDirection();
  FrontLeftWheel.startMotor();

  BackLeftWheel.forwardDirection();
  BackLeftWheel.startMotor();

  delay(TURNING_TIME_MS);
}

void leftTurn()
{
  stopAllMotors();
  FrontLeftWheel.reverseDirection();
  FrontLeftWheel.startMotor();

  BackLeftWheel.reverseDirection();
  BackLeftWheel.startMotor();

  BackRightWheel.forwardDirection();
  BackRightWheel.startMotor();

  FrontRightWheel.forwardDirection();
  FrontRightWheel.startMotor();

  delay(TURNING_TIME_MS);
}

void backwards()
{
  stopAllMotors();

  FrontLeftWheel.reverseDirection();
  FrontLeftWheel.startMotor();

  BackLeftWheel.reverseDirection();
  BackLeftWheel.startMotor();

  BackRightWheel.reverseDirection();
  BackRightWheel.startMotor();

  FrontRightWheel.reverseDirection();
  FrontRightWheel.startMotor();
}

bool checkGPS()
{
  if (Serial1.available() > 0)
  {
    gps.encode(Serial1.read());
    bool isUpdated = gps.location.isUpdated();

#ifdef DEBUG
    if (isUpdated)
    {
      Serial.print("Latitude= ");
      Serial.print(gps.location.lat(), 20);
      Serial.print(" Longitude= ");
      Serial.println(gps.location.lng(), 20);
    }
#endif

    return isUpdated;
  }
  return false;
}

String calculateCardinalDirection(double headingDegrees)
{
  String cardinal;
  if (headingDegrees > 348.75 || headingDegrees < 11.25)
  {
    cardinal = "N";
  }
  else if (headingDegrees > 11.25 && headingDegrees < 33.75)
  {
    cardinal = "NNE";
  }
  else if (headingDegrees > 33.75 && headingDegrees < 56.25)
  {
    cardinal = "NE";
  }
  else if (headingDegrees > 56.25 && headingDegrees < 78.75)
  {
    cardinal = "ENE";
  }
  else if (headingDegrees > 78.75 && headingDegrees < 101.25)
  {
    cardinal = "E";
  }
  else if (headingDegrees > 101.25 && headingDegrees < 123.75)
  {
    cardinal = "ESE";
  }
  else if (headingDegrees > 123.75 && headingDegrees < 146.25)
  {
    cardinal = "SE";
  }
  else if (headingDegrees > 146.25 && headingDegrees < 168.75)
  {
    cardinal = "SSE";
  }
  else if (headingDegrees > 168.75 && headingDegrees < 191.25)
  {
    cardinal = "S";
  }
  else if (headingDegrees > 191.25 && headingDegrees < 213.75)
  {
    cardinal = "SSW";
  }
  else if (headingDegrees > 213.75 && headingDegrees < 236.25)
  {
    cardinal = "SW";
  }
  else if (headingDegrees > 236.25 && headingDegrees < 258.75)
  {
    cardinal = "WSW";
  }
  else if (headingDegrees > 258.75 && headingDegrees < 281.25)
  {
    cardinal = "W";
  }
  else if (headingDegrees > 281.25 && headingDegrees < 303.75)
  {
    cardinal = "WNW";
  }
  else if (headingDegrees > 303.75 && headingDegrees < 326.25)
  {
    cardinal = "NW";
  }
  else if (headingDegrees > 326.25 && headingDegrees < 348.75)
  {
    cardinal = "NNW";
  }
  return cardinal;
}

// https://www.circuitbasics.com/how-to-setup-a-magnetometer-on-the-arduino/
float calculateCurrentDegrees()
{
  compass.getEvent(&event);

  // Calculate the angle of the vector y,x
  float headingDegrees = (atan2(event.magnetic.y, event.magnetic.x) * 180) / M_PI;
  float declinationAngle = 7.816;

  headingDegrees += declinationAngle;

  if (headingDegrees < 0)
  {
    headingDegrees += 360.0;
  }

  Serial.print("Heading: ");
  Serial.println(headingDegrees);
  // delay(250);
  headingDegrees = roundf(headingDegrees * 100000) / 100000; // trim to 5 decimal places
  return headingDegrees;
}

RobotDirection upcomingTurnDirection = RobotDirection::RIGHT;
CardinalDirection currentHeading = CardinalDirection::NORTH;

void acquireDirection(String desired_direction)
{
  int desired_degrees = static_cast<int>(currentHeading);

  int current_degrees = calculateCurrentDegrees();

  while (calculateCardinalDirection(current_degrees) != desired_direction)
  {
    int diff = (desired_degrees - current_degrees) % 360;

    if (diff <= 180)
    {
      leftTurn();
    }
    else
    {
      rightTurn();
    }
  }
  stopAllMotors();
}

static const float grid_height_ft = haversine(SOUTHWEST_LAT, SOUTHWEST_LONG, NORTHEAST_LAT, SOUTHWEST_LONG);
static const float grid_width_ft = haversine(SOUTHWEST_LAT, SOUTHWEST_LONG, SOUTHWEST_LAT, NORTHEAST_LONG);

static const int MAX_NUM_ROWS = round(grid_height_ft / ROBOT_LENGTH);
static const int MAX_NUM_COLUMNS = round(grid_width_ft);
static int gridPos[2];

bool canTurnCurrentDirection()
{
  if (upcomingTurnDirection == RobotDirection::RIGHT)
  {
    if (currentHeading == CardinalDirection::NORTH)
    {
      return gridPos[1] != MAX_NUM_COLUMNS;
    }
    else if (currentHeading == CardinalDirection::SOUTH)
    {
      return gridPos[1] != 0;
    }
  }
  else if (upcomingTurnDirection == RobotDirection::LEFT)
  {
    if (currentHeading == CardinalDirection::NORTH)
    {
      return gridPos[1] != 0;
    }
    else if (currentHeading == CardinalDirection::SOUTH)
    {
      return gridPos[1] != MAX_NUM_COLUMNS;
    }
  }
  return false;
}

bool calculateGridPos()
{
  if (checkGPS()) // if we have a new GPS position, recalculate grid position. Otherwise do nothing.
  {
    const int robot_ft_north = round(haversine(SOUTHWEST_LAT, SOUTHWEST_LONG, gps.location.lat(), SOUTHWEST_LONG));
    const int robot_ft_east = round(haversine(SOUTHWEST_LAT, SOUTHWEST_LONG, SOUTHWEST_LAT, gps.location.lng()));

    const int row_num = robot_ft_north / ROBOT_LENGTH;
    const int col_num = robot_ft_east / ROBOT_WIDTH;
    gridPos[0] = row_num;
    gridPos[1] = col_num;
    return true;
  }
  return false;
}

bool isAboutToBeOutOfBounds() {
  int x = gridPos[0];
  int y = gridPos[1];
  switch (currentHeading) {
    case CardinalDirection::NORTH:
      return x == 1;
    case CardinalDirection::EAST:
      return y == MAX_NUM_COLUMNS;
    case CardinalDirection::SOUTH:
      return x == MAX_NUM_ROWS;
    case CardinalDirection::WEST:
      return y == 1;
    default:
      return false;
  }
}

void loop()
{
  radio.startListening();
  static int curr_servo_degrees = 40; // keeps track of

  while (radio.available())
  {
    #ifdef DEBUG
    Serial.println("Receiving!");
    #endif
    int values[3];
    radio.read(values, sizeof(values));
    int direct_x = values[0];
    int direct_y = values[1];
    int scoop_x = values[2];

    int XPOS = map(direct_x, 2, 1018, 0, 2);
    int YPOS = map(direct_y, 2, 1018, 0, 2);
    #ifdef DEBUG
    Serial.print("X: ");
    Serial.print(XPOS);
    Serial.print(" Y: ");
    Serial.println(YPOS);
    #endif

    int servo_direction = scoop_x;

    if (servo_direction == 2) // go up!
    {
      curr_servo_degrees += SERVO_ADJUSTMENT_FACTOR;
    } else if (servo_direction == 0)
    {
      curr_servo_degrees -= SERVO_ADJUSTMENT_FACTOR;
    }

    curr_servo_degrees = constrain(curr_servo_degrees, 40, 270);

    #ifdef DEBUG
    Serial.print("Curr servo degrees");
    Serial.println(curr_servo_degrees);

    #endif
    // serv.write(75);
    // serv.attach(A0);
    serv.write(curr_servo_degrees);
    // serv.detach();

    if (XPOS == 1 && YPOS == 0)
    {
      #ifdef DEBUG
      Serial.println("Left");
      #endif
      leftTurn();
    }
    // right
    if (XPOS == 1 && YPOS == 2)
    {
      rightTurn();
      #ifdef DEBUG
      Serial.println("Right");
      #endif
    }
    // forward
    if (XPOS == 2 && YPOS == 1)
    {
      fullSteamAhead();
      #ifdef DEBUG
      Serial.println("Forward");
      #endif
    }
    // left
    if (XPOS == 0 && YPOS == 1)
    {
      backwards();
      #ifdef DEBUG
      Serial.println("Backwards");
      #endif
    }
    // topleft
    if (XPOS == 0 && YPOS == 0)
    {
    }
    // topright
    if (XPOS == 2 && YPOS == 0)
    {
    }
    // bottomleft
    if (XPOS == 0 && YPOS == 2)
    {
    }
    // bottomright
    if (XPOS == 2 && YPOS == 2)
    {
    }
    // stop
    if (XPOS == 1 && YPOS == 1)
    {
      stopAllMotors();
    }
    return; // RC mode overrides all
  }

  // GPS MODE IS UNSTABLE, ADJUSTMENT MAY BE REQUIRED
  if(!calculateGridPos())
  {
    return; // GPS not up yet, or we just don't have a reading
  }
  
  if ((gridPos[0] < 0 || gridPos[0] > MAX_NUM_ROWS) || (gridPos[1] < 0 || gridPos[1] > MAX_NUM_COLUMNS))
  {
    lcd.clear();
    lcd.setCursor(0,0);
    lcd.print("Out of bounds");
    lcd.setCursor(1,0);
    lcd.print("Move robot!");
    return;
  }

  switch (currentHeading) { // acquire heading
    case CardinalDirection::NORTH:
      return acquireDirection("N");
    case CardinalDirection::EAST:
      return acquireDirection("E");
    case CardinalDirection::SOUTH:
      return acquireDirection("S");
    case CardinalDirection::WEST:
      return acquireDirection("W");
  }

  while (isAboutToBeOutOfBounds()) // if we're going to be out of bounds, get us a random way back in bounds
  {
    int random_direction = rand() % 5;
    currentHeading = static_cast<CardinalDirection>(random_direction);
  }

  fullSteamAhead(); // proceed forward!
}